/*
 * Copyright 2001-2003 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

#include <malloc.h>
#include "jni.h"
#include "XRGlyphCache.h"
#include "ArrayList.h"
#include "Trace.h"
#include <X11/extensions/Xrender.h>

#define DEFAULT_CACHE_SIZE 1024
#define CACHED_PIXELS 100000
#define TIMES_RENDERED_THRESHOLD 5

#ifndef HEADLESS
XRGlyphCacheInfo* XRGlyphCache_Init(XRenderPictFormat* glyphSetPictFormat)
{
    XRGlyphCacheInfo *cache;
    J2dTraceLn(J2D_TRACE_INFO, "XRGlyphCache_Init");

    cache = (XRGlyphCacheInfo *) malloc(sizeof(XRGlyphCacheInfo));
    if (cache == NULL) {
        J2dRlsTraceLn(J2D_TRACE_ERROR, "XRGlyphCache_Init: could not allocate GlyphCacheInfo");
        return NULL;
    }

    cache->cells = (XRCacheCellInfo *) calloc(DEFAULT_CACHE_SIZE, sizeof(XRCacheCellInfo));
    if(cache->cells == NULL) {
        J2dRlsTraceLn(J2D_TRACE_ERROR, "XRGlyphCache_Init: could not allocate cell space");
        return NULL;
    }

    cache->glyphSet = XRenderCreateGlyphSet (awt_display, glyphSetPictFormat);
    cache->cellCapacity = DEFAULT_CACHE_SIZE;
    cache->usedCells = 0;
    cache->searchPosition = 0;
    cache->pixelsUsed = 0;

    return cache;
}


static jboolean cleanCache(XRGlyphCacheInfo *cache)
{
  J2dTraceLn(J2D_TRACE_INFO, "XRGlyphCache_cleanCache");

  jint i;

  /*First set cold cache entries free*/
  for(i=0; i < cache->cellCapacity; i++) {
    XRCacheCellInfo *cacheCell = &cache->cells[i];
    if(cacheCell->pinned == JNI_FALSE) {
      if(cacheCell->glyphID != 0 && cacheCell->timesRendered < TIMES_RENDERED_THRESHOLD) {
        J2dTraceLn1(J2D_TRACE_INFO, "Removing glyph from cache. ID:", cacheCell->glyphID);
        XRenderFreeGlyphs (awt_display, cache->glyphSet, (Glyph *) &cacheCell->glyphID, 1);
        cacheCell->glyphID = 0;
        cache->pixelsUsed -= (cacheCell->width * cacheCell->height);
        cache->usedCells--;

        if(cacheCell->glyphInfo != NULL) {
	   cacheCell->glyphInfo->cellInfo = NULL;
	}	
      }
    }
     cacheCell->timesRendered = 0;
  }

  cache->searchPosition = 0;
   
  /*If more than 70% are occupied, we enlarge our cache-size by 1.5*/
  jint oldCapacity = cache->cellCapacity;
  XRCacheCellInfo *oldCache = cache->cells;
  jfloat relUsed = ((jfloat) cache->usedCells) / cache->cellCapacity;
  if(relUsed > 0.7)
  {
     cache->cellCapacity = (jint) cache->cellCapacity*1.5;
     J2dTraceLn1(J2D_TRACE_INFO, "Extensing glyph-cache capacity to", cache->cellCapacity);
     
     cache->cells = (XRCacheCellInfo *) realloc(cache->cells, sizeof(XRCacheCellInfo) * (cache->cellCapacity));
     if(cache->cells == NULL) {
        J2dRlsTraceLn(J2D_TRACE_ERROR, "XRGlyphCache_cleanCache: Could not enlarge cell space");
        return JNI_FALSE;
      }
     memset(&cache->cells[oldCapacity], 0, sizeof(XRCacheCellInfo) * (cache->cellCapacity - oldCapacity));

   //If realloc gave us a different location, update pointers in GlyphInfo
   if(oldCache != cache->cells) {
       int cnt;
       for(cnt = 0; cnt < oldCapacity; cnt++) {
           XRCacheCellInfo *cacheCell = &cache->cells[cnt];
           if(cacheCell->glyphID != 0 && cacheCell->glyphInfo != NULL) {
              cacheCell->glyphInfo->cellInfo = cacheCell; /*TODO: we currently mis-use the pointer for our own type*/
           }
        }
    }
  } 

  return JNI_TRUE;
}

static jboolean uploadGlyph(XRGlyphCacheInfo *cache, GlyphInfo *jginfo, XRCacheCellInfo *cacheCell)
{
     J2dTraceLn(J2D_TRACE_INFO, "XRGlyphCache_uploadGlyph");

     unsigned char* glyphBuf = (unsigned char*) jginfo->image; 
     jboolean grayscale = (jginfo->rowBytes == jginfo->width);
     jint dstBytesPerPixel = grayscale ? 1 : 4;
     jint srcBytesPerPixel = grayscale ? 1 : 3;
     jint dstStrideWidth = dstBytesPerPixel * jginfo->width;
     jint srcStrideWidth = srcBytesPerPixel * jginfo->width;
     jint bufLen = dstStrideWidth * jginfo->height;
     jint width = jginfo->width, height = jginfo->height;
     jboolean bufferAllocated = JNI_FALSE;

        if(jginfo->image != NULL) {
          if(grayscale == JNI_TRUE) {

                 /*XRender expects glyph images to be padded to 4-byte boundaries*/
                 if(jginfo->width%4 != 0 || jginfo->image == NULL) { 
                 int pad = 4 - (jginfo->width % 4);
	         dstBytesPerPixel = jginfo->width + pad;
                 bufLen =  dstBytesPerPixel * jginfo->height;
                 glyphBuf = calloc(1, bufLen);
                 bufferAllocated = JNI_TRUE;

                 if(jginfo->image != NULL)  {
	           int line, pix;
	           for(line = 0; line < jginfo->height; line++) {
		     for(pix = 0; pix < jginfo->width; pix++) {
		        glyphBuf[line*jginfo->width + pad*line + pix] = jginfo->image[line*jginfo->width + pix];
		     }
	           }	
                 }
             }

          }else {
            bufLen =  dstStrideWidth * height;
            glyphBuf = calloc(1, bufLen);
            bufferAllocated = JNI_TRUE;

	    int line, srcpix, dstpix;
	    for(line = 0; line < height; line++) {
		srcpix = 0;
		dstpix = 0;
		while(dstpix < dstStrideWidth) {
             /*B*/glyphBuf[line*dstStrideWidth + dstpix + 0] = /*B*/jginfo->image[(line*jginfo->rowBytes + srcpix + 2)];
             /*G*/glyphBuf[line*dstStrideWidth + dstpix + 1] = /*G*/jginfo->image[(line*jginfo->rowBytes + srcpix + 1)];
             /*R*/glyphBuf[line*dstStrideWidth + dstpix + 2] = /*R*/jginfo->image[(line*jginfo->rowBytes + srcpix + 0)];
             /*A*/glyphBuf[line*dstStrideWidth + dstpix + 3] = 255;		
                  srcpix += 3;
                  dstpix += 4;
		}
	    }
	 }
      }else {
         bufLen = 4; //Work arround for bug in older servers
	 glyphBuf = calloc(1, bufLen);
         bufferAllocated = JNI_TRUE;
	 width = 1;
	 height = 1;
      }
     

     Glyph gid = (Glyph) cacheCell->glyphID;
     XGlyphInfo xginfo;
     xginfo.x = (jginfo->topLeftX*-1);
     xginfo.y = (jginfo->topLeftY*-1);
     xginfo.width = width;
     xginfo.height = height;
     xginfo.xOff = cacheCell->xOff;
     xginfo.yOff = cacheCell->yOff;

     XRenderAddGlyphs(awt_display, cache->glyphSet, &gid, &xginfo, 1, glyphBuf, bufLen);
     
     if(bufferAllocated == JNI_TRUE) {
       free(glyphBuf);
     }

     return JNI_TRUE;
}

jboolean XRGlyphCache_CacheGlyphs(XRGlyphCacheInfo *cache, jlong *glyphPtrs, jint glyphCount, jint *ids)
{ 
   jint glyphsStored = 0;
   
   do {
    /**
     * Search for free cache-cells to store glyphs which have not been cached already.
     * The outer loop iterates over all cache-zells, as long as either all have been walked over or all glyphs are stored.
     * The inner loop loops over all non-processed glyphs (if the current zell is free), 
     * and stores it in the free cache-zell if not already cached, and pins it.
     * If the outer-loop finishes before all glyphs are cached, not enough free zells were available and a cache-cleanup is triggered.
     */
      while((cache->searchPosition < cache->cellCapacity) && (glyphsStored < glyphCount)) {
       XRCacheCellInfo *cacheCell = &cache->cells[cache->searchPosition];
       
       while(glyphsStored < glyphCount) {  //loop over unprocessed glyphs
	 GlyphInfo *currGlyph = (GlyphInfo *)jlong_to_ptr(glyphPtrs[glyphsStored]);// &glyphs[glyphsStored];

         if(currGlyph->cellInfo == NULL) { //If the current glyph is not cached, add it to cache if current cacheZell is a free one
           //Evaluate not in loop-head, otherwise we would have to iterate until a free cell
           //even if all glyphs are already cached (common case)
           if(cacheCell->glyphID != 0)
           {
             cache->searchPosition++;
             break;
           }

           currGlyph->cellInfo = cacheCell;
           cacheCell->glyphInfo = currGlyph;
           cacheCell->glyphID = (unsigned short) cache->searchPosition + 1;
           cacheCell->xOff = round(currGlyph->advanceX);
	   cacheCell->yOff = round(currGlyph->advanceY);
           cacheCell->width = currGlyph->width;
           cacheCell->height = currGlyph->height;

           jboolean uploadError = uploadGlyph(cache, currGlyph, cacheCell);
           if(uploadError == JNI_FALSE) {
             return JNI_FALSE;
           }

           cache->pixelsUsed += (currGlyph->width * currGlyph->height);
           cache->usedCells++;
         }

	  //Work that has to be done independently wether the cell was cached or not
          XRCacheCellInfo *usedCell = (XRCacheCellInfo *) currGlyph->cellInfo;
          usedCell->pinned = JNI_TRUE;
          usedCell->timesRendered++;
          glyphsStored++;
        }
     }
       
     //TODO: If we were not able to store all uncached glyphs we free the cache and try again
     if(glyphsStored < glyphCount || cache->pixelsUsed > CACHED_PIXELS) { 
       J2dTraceLn(J2D_TRACE_INFO, "XRGlyphCache: Glyph-cache too small, cleanup triggered");
       if(cleanCache(cache) == JNI_FALSE) {
	  return JNI_FALSE;
       }
     }
   }while(glyphsStored < glyphCount);
 
   /* Remove pin-property from the cache-entries,
    * pass IDs to the caller                     */
   int i;
   for(i=0; i < glyphCount; i++) {
     XRCacheCellInfo *cell = (XRCacheCellInfo*) (((GlyphInfo *)jlong_to_ptr(glyphPtrs[i]))->cellInfo);
     cell->pinned = JNI_FALSE;
     ids[i] = cell->glyphID;
   }

   return JNI_TRUE;
}

#endif /*!Headless*/


// Dev-only debug and trace function.
// void dumpCache(XRGlyphCacheInfo *cache)
// {
//   int i;
//   for(i=0; i < cache->cellCapacity; i++)
//   {
//     XRCacheCellInfo *cell = &cache->cells[i];
//     if(cell->glyphID != 0)
//     {
//       printf("Cache-Position: %d enthaelt glyph: %d mit counter:%d\n", i, cell->glyphID, cell->timesRendered);
//       fflush(stdout);
//     }
//   }
// }
// 	  int line_, pix_;
// 	  for(line_ = 0; line_ < jginfo->height; line_++)
// 	  {
// 		for(pix_ = 0; pix_ < jginfo->rowBytes; pix_+=3)
// 		{
// 		  int r = jginfo->image[line_*jginfo->rowBytes + pix_ + 0];
//                   int g = jginfo->image[line_*jginfo->rowBytes + pix_ + 1];
// 	          int b = jginfo->image[line_*jginfo->rowBytes + pix_ + 2];
//   
//                 if(r > 99)
// 	{
// 	printf("%d,", r);
// 	}else	
//         if(r > 9)
// {
// 	printf(" %d,", r);
// }else
// {
// 	printf("  %d,", r);
// }
// 
//                 if(g > 99)
// 	{
// 	printf("%d,", g);
// 	}else	
//         if(g > 9)
// {
// 	printf(" %d,", g);
// }else
// {
// 	printf("  %d,", g);
// }
// 
//                 if(b > 99)
// 	{
// 	printf("%d,", b);
// 	}else	
//         if(b > 9)
// {
// 	printf(" %d,", b);
// }else
// {
// 	printf("  %d,", b);
// }
// 
// 	printf(" 255, ");
// }
//             printf("\n");
// 	  }	
//         fflush(stdout);
