/*
 * Copyright 2001-2005 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

#include "GlyphImageRef.h"

#ifdef HEADLESS
#include "SurfaceData.h"
#else
#include "X11SurfaceData.h"
#include "XRSurfaceData.h"
#include "GraphicsPrimitiveMgr.h"
#include "glyphblitting.h"
#include "sunfontids.h"
#include "fontscalerdefs.h"
#include "AccelGlyphCache.h"
#include "sunfontids.h"
#include "XRGlyphCache.h"
#include "Trace.h"
#endif /* !HEADLESS */
#include <jlong.h>

#define TEXT_BM_WIDTH   1024
#define TEXT_BM_HEIGHT  32

#define FLOOR_ASSIGN(l, r)\
 if ((r)<0) (l) = ((int)floor(r)); else (l) = ((int)(r))

#ifndef HEADLESS
static XRGlyphCacheInfo* grayCache = NULL;
static XRGlyphCacheInfo* lcdCache = NULL;
#endif

JNIEXPORT void JNICALL
XRT_DrawGlyphList(JNIEnv *env, jobject xtr,
                 jlong dstData, jint numGlyphs, jboolean subPixPos, jboolean rgbOrder, jint lcdContrast,
                 jboolean usePositions, jfloat xOrigin, jfloat yOrigin, jlong *images, jfloat *positions)
{
#ifndef HEADLESS
  J2dTraceLn(J2D_TRACE_INFO, "XRTextRenderer_md: XRT_DrawGlyphList");

  jint i;
  jboolean grayscale = JNI_TRUE;

  X11SDOps *xsdo = (X11SDOps *)jlong_to_ptr(dstData);
  if (xsdo == NULL) {  
      J2dTraceLn(J2D_TRACE_INFO, "XRTextRenderer_md: Destination-Pointer null!");
      return;
    }

  if (numGlyphs > 0) {
      GlyphInfo *ginfo = (GlyphInfo *)jlong_to_ptr(images[0]);
      grayscale = (ginfo->rowBytes == ginfo->width);
      subPixPos = subPixPos && (!grayscale);
  }

  if(grayCache == NULL && grayscale == JNI_TRUE) {
    grayCache = XRGlyphCache_Init(XRenderFindStandardFormat(awt_display, PictStandardA8));
  }else
  if(lcdCache == NULL && grayscale == JNI_FALSE) {
    lcdCache = XRGlyphCache_Init(XRenderFindStandardFormat(awt_display, PictStandardARGB32));
  }

   int eltCnt = -1;
   jint *ids; 
   XGlyphElt32 *elts;

   /*for less then 128 glyphs we allocate temorary mem on stack*/
   jint idArray[128];
   XGlyphElt32 eltArray[128];
   if(numGlyphs <= 128) {
     ids = &idArray[0];
     elts = &eltArray[0];
   }else {
     ids = (jint *) malloc(numGlyphs * sizeof(jint));
     elts = (XGlyphElt32 *) malloc(numGlyphs * sizeof(XGlyphElt32));
   }

   float advX = xOrigin; //xOrigin;/*Contains all advance-corrections summed up*/
   float advY = yOrigin; //yOrigin; 
   int oldPosX = 0, oldPosY = 0; /*Stored to be able to caluclate integer-advance to last glyph*/

   if(subPixPos == JNI_TRUE) {
     advX += 0.1666667f;
     advY += 0.1666667f;
   }else {
     advX += 0.5f;
     advY += 0.5f;
  }

   XRGlyphCacheInfo *glyphCache = (grayscale == JNI_TRUE) ?  grayCache : lcdCache;
   jboolean cacheResult = XRGlyphCache_CacheGlyphs(glyphCache, images, numGlyphs, ids);
   if(cacheResult == JNI_FALSE){
      J2dTraceLn(J2D_TRACE_INFO, "XRTextRenderer_md: Glyph caching failed");
      return;
   }

   for(i=0; i < numGlyphs; i++)
   {
      GlyphInfo *jginfo = (GlyphInfo *)jlong_to_ptr(images[i]);
      XRCacheCellInfo *cellInfo = (XRCacheCellInfo *) jginfo->cellInfo;

	/** If the cached and the required relative position is equal (implies integer value), we can reuse
	 *  the existing ELT, and let XRender do the positioning.
	 *  If it does not match the positioning information stored on the server, create a new ELT.
	 *  advX/Y and posX/Y contain all advance-corrections summed up, so that we don't forget rounding errors happend in the past.
         *  The actual advance is computed from the difference to the last integer-correction-position.
	 */
	int posX = 0, posY = 0;
	if(usePositions || (jginfo->advanceX != ((float)cellInfo->xOff) || jginfo->advanceY != ((float)cellInfo->yOff)) || eltCnt == -1) 
        {
          /*Setup the next elt*/
	  eltCnt++; 
	  elts[eltCnt].glyphset = glyphCache->glyphSet;
          elts[eltCnt].chars = (unsigned int*) &ids[i];
	  elts[eltCnt].nchars = 1;

  	  if(usePositions)
	  {
            /*In this case advX only stores rounding errors*/
            jfloat x = positions[i*2] + advX;
            jfloat y = positions[i*2 + 1] + advY;
            FLOOR_ASSIGN(posX, x);
            FLOOR_ASSIGN(posY, y);
 	    advX -= cellInfo->xOff;
 	    advY -= cellInfo->yOff;
          }else
          {
	    /* 
             * Calculate next glyph's position in the case of relative positioning.
             * In XRender we can only position glyphs using integer coordinates, therefor
             * we sum all the advances up as float, and convert them to integer later.
             * This way rounding-error can be corrected,
             * and is required to be consistent with the software loops.
             */
             FLOOR_ASSIGN(posX, advX);
             FLOOR_ASSIGN(posY, advY);
             //Advance of ELT = difference between stored relative positioning information and required float.
	     advX += (jginfo->advanceX - cellInfo->xOff);
	     advY += (jginfo->advanceY - cellInfo->yOff);
          }

            /*Offset of the current glyph is the difference to the last glyph and this one*/
	    elts[eltCnt].xOff = (posX - oldPosX);
	    elts[eltCnt].yOff = (posY - oldPosY);

	    oldPosX = posX;
	    oldPosY = posY;
	}else
	{
           elts[eltCnt].nchars++;
	}
   }

  /*TODO: Also integrate into MaskBuffer??*/
  XRenderCompositeText32(awt_display, PictOpOver, xrSrcData.src, xsdo->xrPic, None, 0, 0, 0, 0, elts, eltCnt+1);

  if(numGlyphs > 128) {
    free(ids);
    free(elts);
  }
#endif
}

